import json
from dataclasses import dataclass

import requests
from dataset_sh.multipart.multipart import MultipartFileWriter


@dataclass
class DirectDownloadInstruction:
    direct_download_url: str
    type: str = 'direct'


@dataclass
class MultipartDownloadInstruction:
    file_length: int
    checksum: str
    multipart_download_url: str
    type: str = 'multipart'


class MultipartDownloadClient(object):

    def __init__(self, source_url, target):
        self.target = target
        self.source_url = source_url

    def _get_instruction(self):
        resp = requests.get(self.source_url, params={'action': 'request-info'})
        resp.raise_for_status()
        resp_json = json.loads(resp.content.decode('utf-8'))
        download_type = resp_json.get('type', None)
        if download_type == 'direct':
            return DirectDownloadInstruction(direct_download_url=resp_json['direct_download_url'])
        elif download_type == 'multipart':
            return MultipartDownloadInstruction(
                file_length=resp_json['file_length'],
                checksum=resp_json['checksum'],
                multipart_download_url=resp_json['multipart_download_url']
            )
        else:
            raise ValueError(f'Unknown instruction type: {download_type}')

    def download(self):
        instruction = self._get_instruction()
        if instruction.type == 'direct':
            resp = requests.get(instruction.direct_download_url)
            resp.raise_for_status()
            with open(self.target, 'wb') as fd:
                fd.write(resp.content)

        else:
            writer = MultipartFileWriter(
                self.target,
                file_length=instruction.file_length,
                checksum=instruction.checksum,
            )
            writer.start()
            todo = writer.get_uncompleted_parts()
            for part_id in todo:
                part_file = writer.part_file(part_id)

                resp = requests.get(instruction.multipart_download_url, params={'part_id': part_id})
                resp.raise_for_status()
                with open(part_file, 'wb') as fd:
                    fd.write(resp.content)

                writer.mark_file_as_complete(part_id)
            writer.done()



import warnings


def to_dataframe(collection: 'CollectionReader'):
    try:
        import pandas as pd
    except ModuleNotFoundError as e:
        warnings.warn('cannot load module pandas, maybe it is not installed ?', UserWarning)
        return None

    return pd.DataFrame(collection)



import os.path

CONFIG_JSON = os.path.expanduser('~/.dataset_sh_config.json')
STORAGE_BASE = os.path.expanduser('~/dataset_sh/storage')
ALIAS_FILE = os.path.expanduser('~/.dataset.sh.host-alias.json')
DSH_DEFAULT_HOST = os.getenv('DSH_DEFAULT_HOST', 'https://repo.dataset.sh')

SAMPLE_CHAR_COUNT = 1000 * 10

DEFAULT_COLLECTION_NAME = 'main'



#!/usr/bin/env python3
import os.path
from dataclasses import dataclass
from getpass import getpass
from urllib.parse import urlparse

import click

from dataset_sh.server.core import RepoServerConfig


@dataclass
class RepoServerConfigCtx:
    path: str
    config: RepoServerConfig
    is_new: bool = False


@click.group(name='config')
@click.option(
    '--config', '-c',
    envvar='DATASET_SH_SERVER_CONFIG_FILE',
    help='location of server config file.',
    default='./dataset-sh-server-config.json',
    type=click.Path()
)
@click.pass_context
def config_cli(ctx, config):
    """config dataset.sh server"""
    if os.path.exists(config):
        ctx.obj = RepoServerConfigCtx(path=config, config=RepoServerConfig.load_from_file(config))
    else:
        ctx.obj = RepoServerConfigCtx(path=config, config=RepoServerConfig(), is_new=True)


@click.command(name='set-password')
@click.argument('username')
@click.pass_obj
def set_password(ctx, username):
    """set password for user, and generate a new access key"""
    password = getpass('Password:')
    ctx.config.update_password(username, password)
    ctx.config.write_to_file(ctx.path)
    key = ctx.config.generate_key(username, password)
    click.echo(f'Access Key: {key}\n')
    click.echo(f'Add this key to your client with this command: ')
    click.echo(f'  dataset.sh profile add -h {ctx.config.hostname}')


@click.command(name='init')
@click.pass_obj
def init_config(ctx):
    """Create an empty config file"""
    ctx.config.write_to_file(ctx.path)


@click.command(name='set-hostname')
@click.argument('hostname')
@click.pass_obj
def set_hostname(ctx, hostname):
    """set server hostname"""
    if not is_valid_url(hostname):
        click.echo(f'invalid hostname {hostname}')
        raise click.Abort()
    ctx.config.hostname = hostname
    ctx.config.write_to_file(ctx.path)


@click.command(name='upload')
@click.argument(
    'action',
    type=click.Choice(['disable', 'enable'], case_sensitive=False))
@click.pass_obj
def config_upload_permission(ctx, action):
    """config upload permission"""
    if action == 'enable':
        ctx.config.allow_upload = True
    elif action == 'disable':
        ctx.config.allow_upload = False
    else:
        click.echo(f'Unknown action {action}.')
        raise click.Abort()

    ctx.config.write_to_file(ctx.path)


@click.command(name='read-access')
@click.argument(
    'mode',
    type=click.Choice(['public', 'private'], case_sensitive=False))
@click.pass_obj
def config_read_access(ctx, mode):
    """
    config read accessibility for the server:

        * `dataset-sh-server config read-access public`:
            allow public read on server.

        * `dataset-sh-server config read-access private`:
            only logged-in user can read content on this server.

    """
    if mode == 'public':
        ctx.config.require_auth = False
    elif mode == 'private':
        ctx.config.require_auth = True
    else:
        click.echo(f'Unknown mode {mode}.')
        raise click.Abort()

    ctx.config.write_to_file(ctx.path)


def is_valid_url(url):
    parsed_url = urlparse(url)
    return parsed_url.scheme in ['http', 'https']


config_cli.add_command(set_password)
config_cli.add_command(init_config)
config_cli.add_command(set_hostname)
config_cli.add_command(config_upload_permission)
config_cli.add_command(config_read_access)

if __name__ == '__main__':  # pragma: no cover
    config_cli()



import click

from dataset_sh.remote_utils.profile import DatasetClientProfileConfig, ask_hostname, ask_name, ask_access_key, \
    ask_override
import dataset_sh.constants as DatasetConstants


@click.group(name='profile')
def profile_cli():
    """to manage dataset.sh remote profiles"""
    pass


@profile_cli.command(name='add')
@click.option('--name', '-n', 'name', help='profile name.', default=None)
@click.option('--host', '-h', 'host', help='host address or alias.', default=None)
@click.option('--force', '-f', 'force',
              is_flag=True, help='force override existing profile.',
              default=False)
def add_profile(name, host, force):
    """
    Add access key to dataset.sh client.
    """
    cfg = DatasetClientProfileConfig.load_profiles()
    if host is None or host.strip() == '':
        host = ask_hostname('https://repo.dataset.sh')

    if name is None:
        name = ask_name()

    if cfg.get_profile(name):
        click.echo("")
        click.echo('Find existing profile with the same name.')
        if not force:
            if not ask_override():
                raise ValueError(f'Profile {name} exists.')

    key = ask_access_key()

    cfg.set_profile(host, key, name=name)
    cfg.save()

    click.echo(f'Profile saved in {DatasetConstants.CONFIG_JSON}')


@profile_cli.command(name='list')
def list_profile():
    """
    List remote profiles.
    """
    cfg = DatasetClientProfileConfig.load_profiles()
    for p in cfg.profiles:
        if p.name:
            click.echo(f"Profile: {p.name}")
        click.echo(f"Host: {p.host}")
        click.echo('\n')


@profile_cli.command(name='print')
@click.argument('name')
def print_profile(name):
    """
    print remote profile information.
    """
    profile = DatasetClientProfileConfig.load_profiles().get_profile(name)
    if profile:
        click.echo(f'Profile: {profile.name}')
        click.echo(f'Host: {profile.host}')
    else:
        raise ValueError(f'profile {name} does not exist')


@profile_cli.command(name='remove')
@click.argument('name')
def remove_profile(name):
    """
    Remove existing profile.
    """
    DatasetClientProfileConfig.load_profiles().remove_profile(name).save()



import json
import re
from typing import Optional

from dataset_sh.clients.locator import DatasetLocator
from dataset_sh.dependency.model import DatasetDependencies, DatasetDependencyHostGroup, DatasetDependencyItem


def parse(content):
    ret = DatasetDependencies(**json.loads(content))
    for hg in ret.dependencies:
        for item in hg.datasets:
            DatasetLocator.from_str(item.source)
            if item.target:
                DatasetLocator.from_str(item.target)
    return ret


def parse_file(fp: str) -> DatasetDependencies:
    with open(fp) as f:
        return parse(f.read())

# class ParsingError(ValueError):
#     def __init__(self, msg):
#         super().__init__(msg)
#
#
# def parse(lines: str) -> DatasetDependencies:
#     result = DatasetDependencies()
#     current_host_group = DatasetDependencyHostGroup(host=DatasetConstants.CENTRAL_HOST)
#     for line in lines.split('\n'):
#         if is_host_line(line):
#             if len(current_host_group.datasets) > 0:
#                 result.datasets_by_host.append(current_host_group)
#             current_host_group = DatasetDependencyHostGroup(host=parse_host(line))
#         else:
#             item = parse_item(line)
#             if item is not None:
#                 current_host_group.datasets.append(item)
#     if len(current_host_group.datasets) > 0:
#         result.datasets_by_host.append(current_host_group)
#     return result
#
#
# def is_host_line(line: str):
#     return line.lower().startswith('host:')
#
#
# def parse_host(line: str):
#     if line.lower().startswith('host:'):
#         return line[5:].strip()
#     return None
#
#
# def is_valid_name(name):
#     pattern = r"^[a-zA-Z][a-zA-Z0-9_-]*/[a-zA-Z][a-zA-Z0-9_-]*$"
#     if re.match(pattern, name):
#         return True
#     return False
#
#
# def parse_kw_arg(kw_arg):
#     if '=' in kw_arg:
#         parts = kw_arg.split('=')
#         if len(parts) == 2:
#             k, w = parts
#             k = k.lower().strip()
#             if k in ['version', 'tag']:
#                 return k, w.strip()
#     else:
#         return None
#
#
# def parse_item(line: str) -> Optional[DatasetDependencyItem]:
#     line = line.strip()
#
#     if len(line) == 0 or line.startswith('#'):
#         return None
#
#     if '?' in line:
#         parts = line.split('?')
#         if len(parts) == 2:
#             dataset_fullname, kw_args = parts
#             item = DatasetDependencyItem(name=dataset_fullname.strip())
#             for kw_arg in kw_args.split('&'):
#                 kw = parse_kw_arg(kw_arg)
#                 if kw:
#                     k, w = parse_kw_arg(kw_arg)
#                     if k == 'version':
#                         item.version = w
#                     elif k == 'tag':
#                         item.tag = w
#             return item
#         raise ParsingError(f"Cannot parse dataset declaration: {line}")
#     else:
#         return DatasetDependencyItem(name=line)



def generate_download_code_py(hostname, dataset_name):
    template = f'''
import dataset_sh as dsh
dsh.remote(host="{hostname}").dataset("{dataset_name}").latest().download()
    '''.strip()
    return template


def generate_download_code_py_with_version(hostname, dataset_name, version):
    template = f'''
import dataset_sh as dsh
dsh.remote(host="{hostname}").dataset("{dataset_name}").version("{version}").download()
    '''.strip()
    return template


def generate_download_code_py_with_tag(hostname, dataset_name, tag):
    template = f'''
import dataset_sh as dsh
dsh.remote(host="{hostname}").dataset("{dataset_name}").tag("{tag}").download()
    '''.strip()
    return template


def generate_download_code_bash(hostname, dataset_name):
    template = f'''
dataset.sh remote -h {hostname} download {dataset_name}
    '''.strip()
    return template


def generate_download_code_bash_with_version(hostname, dataset_name, version):
    template = f'''
dataset.sh remote -h {hostname} download {dataset_name} -v {version}
    '''.strip()
    return template


def generate_download_code_bash_with_tag(hostname, dataset_name, tag):
    template = f'''
dataset.sh remote -h {hostname} download {dataset_name} -t {tag}
    '''.strip()
    return template


def generate_download_code(language, hostname, dataset_name, version=None, tag=None):
    if language == 'bash':
        if version:
            return generate_download_code_bash_with_version(hostname, dataset_name, version)
        elif tag and tag != 'latest':
            return generate_download_code_bash_with_tag(hostname, dataset_name, tag)
        else:
            return generate_download_code_bash(hostname, dataset_name)
    if language == 'python':
        if version:
            return generate_download_code_py_with_version(hostname, dataset_name, version)
        elif tag and tag != 'latest':
            return generate_download_code_py_with_tag(hostname, dataset_name, tag)
        else:
            return generate_download_code_py(hostname, dataset_name)



import hashlib
import json
import os
from datetime import datetime

import requests


def checksum(file_path):
    md5_hash = hashlib.sha256()

    with open(file_path, "rb") as file:
        while True:
            data = file.read(65536)
            if not data:
                break
            md5_hash.update(data)

    return md5_hash.hexdigest()


def filesize(file_path):
    file_size = os.path.getsize(file_path)
    return file_size


def file_creation_time(file_path):
    creation_time = os.path.getctime(file_path)
    # Convert it to a readable format
    readable_time = datetime.fromtimestamp(creation_time)
    return readable_time


def read_url_json(url, headers):  # pragma: no cover
    response = requests.get(url, headers=headers)
    response.raise_for_status()
    return response.json()


def download_url(url, target, headers):
    response = requests.get(url, headers=headers, stream=True)

    has_tqdm = True
    try:
        from tqdm import tqdm
    except ModuleNotFoundError:  # pragma: no cover
        has_tqdm = False

    if response.status_code == 200:
        with open(target, "wb") as file:
            if has_tqdm:
                file_size = int(response.headers.get('content-length', 0))
                desc = "(Unknown total file size)" if file_size == 0 else ""
                with tqdm(
                        total=file_size, unit='iB', unit_scale=True
                ) as bar:
                    for data in response.iter_content(chunk_size=1024):
                        size = file.write(data)
                        bar.update(size)
            else:
                for data in response.iter_content(chunk_size=1024):  # pragma: no cover
                    size = file.write(data)

        response.close()
    else:
        server_err = json.loads(response.content.decode('utf-8'))
        reason = server_err.get('reason', '')
        e = ValueError(f"Failed to download the file. Status code: {response.status_code}, reason: {reason}")
        response.close()
        raise e


def download_if_not_exists(url, filepath=None, **kwargs):
    """
    Download file to filepath if filepath not exists, additional arguments will be passed to `requests.get`.
    if filepath is None, the file will be downloaded to a file at current working dir with the
    last path element of the given url.

    Args:
        url:
        filepath:

    Returns:
        the downloaded file path.

    """
    if filepath is None:
        filename = url.split('/')[-1]
        if not filename:
            filename = '__download'
        filepath = os.path.join(os.getcwd(), filename)

    # Check if the file already exists
    if not os.path.exists(filepath):
        try:
            # The file does not exist; download it
            print(f"Downloading file from {url} to {filepath}")
            response = requests.get(url, **kwargs)
            response.raise_for_status()  # Raises HTTPError, if one occurred

            # Write the downloaded content to a file
            with open(filepath, 'wb') as file:
                file.write(response.content)
        except Exception as e:
            print(f"Failed to download or write file: {e}")
            # If the file was partially written, remove it to clean up
            if os.path.exists(filepath):
                os.remove(filepath)
                print(f"Removed incomplete file: {filepath}")
            # Re-raise the exception to indicate failure
            raise
    else:
        print(f"File already exists: {filepath}")

    return filepath



import json
from typing import Optional

import click
from click import BadParameter

from dataset_sh.clients.obj import LocalStorage
from dataset_sh.constants import DEFAULT_COLLECTION_NAME
from dataset_sh.utils.misc import parse_tags
from dataset_sh.utils.usage.read_data_codegen import get_read_data_code


def _get_dataset_version(ctx, name, version, tag):
    dataset = ctx.local_storage.dataset(name)
    if version:
        dv = dataset.version(version)
    elif tag:
        dv = dataset.tag(tag)
    else:
        dv = dataset.latest()
    return dv


class LocalCtx(object):

    def __init__(self, folder: Optional[str]):
        self.folder = folder
        self.local_storage = LocalStorage(folder)


@click.group(name='local')
@click.option('--folder', '-f', 'folder', help='location of the dataset storage folder', default=None)
@click.pass_context
def local_cli(ctx, folder):
    """to interact with dataset.sh local storage"""
    ctx.obj = LocalCtx(folder)


@local_cli.command(name='import-as-latest')
@click.argument('name')
@click.argument('file')
@click.option('--tag', '-t', default='', help='additional tags')
@click.pass_obj
def import_as_latest(ctx, name, file, tag):
    """import dataset from file, and automatically tag it as latest"""
    click.echo(f'importing local file from {file}')
    ctx.local_storage.dataset(name).import_file(file_path=file, tags=parse_tags(tag))


@local_cli.command(name='import')
@click.argument('name')
@click.argument('file')
@click.option('--tag', '-t', default='', help='version tags')
@click.pass_obj
def import_(ctx, name, file, tag):
    """import dataset from file and do not tag"""
    click.echo(f'importing local file from {file}')
    ctx.local_storage.dataset(name).import_file(file_path=file, tags=parse_tags(tag), as_latest=False)


@local_cli.command(name='meta')
@click.argument('name')
@click.option('--version', '-v', help='use this option to select version', default=None)
@click.option('--tag', '-t', help='use this option to select version with tag', default='latest')
@click.pass_obj
def print_meta(ctx, name, version, tag):
    """print metadata of a dataset"""
    dv = _get_dataset_version(ctx, name, version, tag)
    if dv.exists():
        meta = dv.meta()
        if meta:
            click.echo(json.dumps(meta, indent=2))
    else:
        raise BadParameter(f'{dv} do not exists.')


@local_cli.command(name='list-collections')
@click.argument('name')
@click.option('--version', '-v', help='use this option to select version', default=None)
@click.option('--tag', '-t', help='use this option to select version with tag', default='latest')
@click.pass_obj
def list_collections(ctx, name, version, tag):
    """
    list collections of a dataset
    """
    dv = _get_dataset_version(ctx, name, version, tag)
    if dv.exists():
        meta = dv.meta_object()
        click.echo(f'Total collections: {len(meta.collections)}')
        for coll in meta.collections:
            click.echo(coll.name)
    else:
        raise BadParameter(f'{dv} do not exists.')


@local_cli.command(name='code-usage')
@click.argument('name')
@click.option('--collection', '-c', help='collection name', default=DEFAULT_COLLECTION_NAME)
@click.option('--version', '-v', help='use this option to select version', default=None)
@click.option('--tag', '-t', help='use this option to select version with tag', default='latest')
@click.pass_obj
def print_code(ctx, name, collection, version, tag):
    """print data model code of a dataset collection"""
    dv = _get_dataset_version(ctx, name, version, tag)
    if dv.exists():

        if not dv.collection_exists(collection):
            click.echo(f'collection {collection} does not exist. Found the following collection in dataset {name}')
            for c in dv.list_collection_names():
                click.echo(f'  {c}')
            click.echo('---')
            raise BadParameter(f'collection {collection} does not exist.')
        reader_code = get_read_data_code(name, collection, version=version, tag=tag)
        click.echo(reader_code)
    else:
        raise BadParameter(f'{dv} do not exists.')


@local_cli.command(name='print-sample')
@click.argument('name')
@click.option('--collection', '-c', help='collection name', default=DEFAULT_COLLECTION_NAME)
@click.option('--version', '-v', help='use this option to select version', default=None)
@click.option('--tag', '-t', help='use this option to select version with tag', default='latest')
@click.pass_obj
def print_sample(ctx, name, collection, version, tag):
    """print sample content of a dataset collection """
    dv = _get_dataset_version(ctx, name, version, tag)
    if dv.exists():
        if not dv.collection_exists(collection):
            click.echo(f'collection {collection} does not exist. Found the following collection in dataset {name}')
            for c in dv.list_collection_names():
                click.echo(f'  {c}')
            click.echo('---')
            raise BadParameter(f'collection {collection} does not exist.')

        samples = dv.sample(collection)
        click.echo(json.dumps(samples, indent=2))

    else:
        raise BadParameter(f'{dv} do not exists.')


@local_cli.command(name='remove')
@click.argument('name')
@click.option('--force', '-f', default=False, help='Force remove dataset without confirmation.', is_flag=True)
@click.pass_obj
def remove(ctx, name, force):
    """remove all versions of a managed dataset"""
    dataset = ctx.local_storage.dataset(name)
    do_remove = False
    if force:
        do_remove = True
    else:
        confirmation = click.prompt(f'Are you sure you want to remove all versions of dataset {name}? (y/N): ')
        if confirmation.lower() == 'y':
            do_remove = True

    if do_remove:
        dataset.delete()


@local_cli.command(name='remove-version')
@click.argument('name')
@click.option('--version', '-v', help='use this option to select version', default=None)
@click.option('--tag', '-t', help='use this option to select version with tag', default=None)
@click.option('--force', '-f', default=False, help='Force remove dataset without confirmation.', is_flag=True)
@click.pass_obj
def remove_version(ctx, name, version, tag, force):
    """remove a version from a managed dataset"""
    dv = _get_dataset_version(ctx, name, version, tag)
    do_remove = False
    if force:
        do_remove = True
    else:
        confirmation = click.prompt(f'Are you sure you want to remove all versions of dataset {name}? (y/N): ')
        if confirmation.lower() == 'y':
            do_remove = True

    if do_remove:
        dv.delete()


@local_cli.command(name='list')
@click.option('--namespace', '-n', help='select dataset store space to list.', default=None)
@click.pass_obj
def list_datasets(ctx, namespace):
    """list datasets"""
    if namespace:
        items = ctx.local_storage.namespace(namespace).datasets()
    else:
        items = ctx.local_storage.datasets()

    click.echo(f'\nFound {len(items)} datasets:\n')
    items = sorted(items, key=lambda x: f'{x.namespace}/{x.dataset_name}')
    for item in items:
        click.echo(f'  {item.namespace}/{item.dataset_name}')
    click.echo('')


@local_cli.command(name='list-version')
@click.argument('name')
@click.pass_obj
def list_dataset_versions(ctx, name):
    """list dataset versions"""
    dataset = ctx.local_storage.dataset(name)
    versions = dataset.versions()

    click.echo(f'\nFound {len(versions)} versions:\n')
    for item in versions:
        click.echo(f'  {item.version}')
    click.echo('')


@local_cli.command(name='tag')
@click.argument('name')
@click.argument('tag')
@click.argument('version')
@click.pass_obj
def tag_dataset_version(ctx, name, tag, version):
    """Tag dataset version"""
    dataset = ctx.local_storage.dataset(name)
    dataset.set_tag(tag=tag, version=version)


@local_cli.command(name='untag')
@click.argument('name')
@click.argument('tag')
@click.pass_obj
def untag_dataset_version(ctx, name, tag):
    """Remove dataset version tag"""
    dataset = ctx.local_storage.dataset(name)
    dataset.remove_tag(tag)


@local_cli.command(name='tag-info')
@click.argument('name')
@click.argument('tag')
@click.pass_obj
def print_tag_info(ctx, name, tag):
    """Print dataset version tag information"""
    dataset = ctx.local_storage.dataset(name)
    tagged_version = dataset.resolve_tag(tag=tag)
    if tagged_version:
        click.echo(f"{tag} : {tagged_version}")
    else:
        click.echo(f"{tag} do not exists")



import os

from flask import Flask, request, jsonify

from dataset_sh.multipart.multipart import MultipartFileWriter

app = Flask(__name__)

upload_base = '/tmp/upload-test'
os.makedirs(upload_base, exist_ok=True)


@app.route("/upload/<filename>", methods=['POST'])
def hello_world(filename):
    checksum_value = request.args['checksum_value']
    file_length = int(request.args['file_length'])
    action = request.args['action']
    target_file = os.path.join(upload_base, filename)

    writer = MultipartFileWriter(
        target_file,
        checksum=checksum_value,
        file_length=file_length
    )

    if action == 'init':
        existing = writer.load_existing_progress()
        if existing is None:
            writer.start()
            existing = writer.load_existing_progress()
        return jsonify(existing.model_dump(mode='json'))
    elif action == 'upload':
        part_id = request.args['part_id']
        part_file = writer.part_file(int(part_id))
        with open(part_file, 'wb') as out:
            out.write(request.data)
        return '', 204
    elif action == 'done':
        writer.done()
        return '', 204
    else:
        return '', 400


if __name__ == '__main__':
    app.run(port=9900, debug=True)
